/*
 * Capabilities jabber support (XEP-0115).
 *
 * Grigory Holomiev (omever@gmail.com) / 2008
 */

#include "Caps.h"
#include <stdbool.h>

#ifdef _FREEBSD7
typedef int _Bool;
#endif

#include <mhash.h>

#ifdef __cplusplus
#include <algorithm>
extern "C" {
#endif

void* caps_init(int version)
{
	Caps *c = new Caps();

	c->setNode("http://jit.mytlt.ru");
	Identity tmp;
	switch(version) {
	case CAPS_NULL:
		break;
	case CAPS_CONTACT:
		tmp.category = "client";
		tmp.type = "pc";
		tmp.name = "JIT Contact part";
		tmp.lang = "ru";

		c->addIdentity(tmp);
		c->addFeature(NS_VCARD);
		c->addFeature(NS_LAST);
		c->addFeature(NS_CAPS);
		c->addFeature(NS_CAPS_TUNE);
		c->addFeature(NS_CAPS_MOOD);
		c->addFeature(NS_CAPS_ACTIVITY);
		break;
	case CAPS_TRANSPORT:
		tmp.lang = "ru";
		tmp.category = "gateway";
		tmp.name = "Jabber ICQ Transport";
		tmp.type = "icq";

		c->addIdentity(tmp);

		c->addFeature(NS_REGISTER);
		c->addFeature(NS_SEARCH);
		c->addFeature(NS_VERSION);
		c->addFeature(NS_TIME);
		c->addFeature(NS_GATEWAY);
		c->addFeature(NS_VCARD);
		c->addFeature(NS_LAST);
		c->addFeature(NS_CAPS);
		c->addFeature(NS_CAPS_MOOD);
		c->addFeature(NS_CAPS_MOOD_N);
		c->addFeature(NS_CAPS_TUNE);
		c->addFeature(NS_CAPS_TUNE_N);
		c->addFeature(NS_CAPS_ACTIVITY);
		c->addFeature(NS_CAPS_ACTIVITY_N);

		break;
	default:
		break;
	}

	return (void*)c;
}

void caps_deinit(void*c)
{
	delete((Caps*)c);
}

void caps_add_identity(void*caps, const char *category, const char *type, const char *lang,const char *name)
{
	Caps *c = (Caps*)caps;
	Identity tmp;
	tmp.lang = std::string(lang);
	tmp.category = std::string(category);
	tmp.name = std::string(name);
	tmp.type = std::string(type);

	c->addIdentity(tmp);
}

void caps_add_feature(void*caps, const char *feature)
{
	Caps *c = (Caps*)caps;
	c->addFeature(std::string(feature));
}

void caps_fill_info(void *caps, xmlnode x) {
		Caps *c = (Caps*)caps;
		c->fill_info(x);
}

void caps_fill_string(void *caps, char *dest, int len, const char *version)
{
	Caps *c = (Caps*)caps;
	std::string res = c->to_string(version);
	strncpy(dest, res.c_str(), len-1);
}

const char * caps_get_node(void *caps)
{
	Caps *c = (Caps*)caps;
	return c->getNode().c_str();
}

#ifdef __cplusplus
};
#endif

Caps::Caps()
{
	_identities.clear();
	_features.clear();
}

Caps::~Caps()
{
}

void Caps::init()
{

}

std::string Caps::b64(std::string in)
{
	char alphabet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

	std::string out("");
	size_t sz = in.size();
	unsigned char tmp;
	for(size_t i=0; i<sz; ++i) {
	    unsigned char c = in[i];
	    if(i%3 == 0) {
	        out += alphabet[c>>2 & 63];
	        tmp = (c & 3) << 4;
	    }
	    if(i%3 == 1) {
	        out += alphabet[(tmp | (c>>4)&15) & 63];
	        tmp = (c&15) << 2;
	    }
	    if(i%3 == 2) {
	        out += alphabet[(tmp | (c>>6)&3) & 63];
	        out += alphabet[c & 63];
	        tmp = 0;
	    }
	}
	if(sz % 3 == 1) {
	    out += alphabet[tmp];
	    out += "==";
	}
	if(sz % 3 == 2) {
	    out += alphabet[tmp];
	    out += "=";
	}

	return out;
}

std::string Caps::to_string(const char *version)
{
	sort(_features.begin(), _features.end());
	sort(_identities.begin(), _identities.end());

	std::string S("");

	std::vector<Identity>::iterator i, iend = _identities.end();
	for(i=_identities.begin(); i!=iend; ++i) {
		if(version != NULL) 
			S += i->category + "/" + i->type + "/" + i->lang + "/" + version + "<";
		else
			S += i->category + "/" + i->type + "/" + i->lang + "/" + i->name + "<";
	}

	std::vector<std::string>::iterator j, jend = _features.end();
	for(j=_features.begin(); j!=jend; ++j) {
		S += *j + "<";
	}
	MHASH td = mhash_init(MHASH_SHA1);
	if(td == MHASH_FAILED) {
		return "";
	}

	size_t sz = mhash_get_block_size(MHASH_SHA1);
	if(!sz) {
		return "";
	}

	char *buffer = new char [sz];
	mhash(td, S.c_str(), S.length());
	mhash_deinit(td, buffer);

	S=b64(std::string(buffer));

	return S;
}

void Caps::fill_info(xmlnode &query)
{
	xmlnode info;
	std::vector<Identity>::iterator i, iend = _identities.end();
	for(i=_identities.begin(); i!=iend; ++i) {
	    info = xmlnode_insert_tag(query, "identity");
	    xmlnode_put_attrib(info,"category", i->category.c_str());
	    xmlnode_put_attrib(info,"type", i->type.c_str());
		xmlnode_put_attrib(info,"name", i->name.c_str());
		xmlnode_put_attrib(info,"lang", i->lang.c_str());
	}

	std::vector<std::string>::iterator j, jend = _features.end();
	for(j=_features.begin(); j!=jend; ++j) {
	    info = xmlnode_insert_tag(query,"feature");
	    xmlnode_put_attrib(info,"var",j->c_str());
	}
}

void Caps::addIdentity(Identity i)
{
	_identities.push_back(i);
}

void Caps::addFeature(std::string f)
{
	_features.push_back(f);
}

void Caps::setNode(std::string newnode)
{
	_node = newnode;
}

void Caps::clearIdentities()
{
	_identities.clear();
}

void Caps::clearFeatures()
{
	_features.clear();
}

std::string& Caps::getNode()
{
	_cnode = _node + to_string();
	return _cnode;
}
