/*
 * Capabilities jabber support (XEP-0115).
 * 
 * Grigory Holomiev (omever@gmail.com) / 2008
 */
#ifndef CAPS_H_
#define CAPS_H_

#include "icqtransport.h"

#define CAPS_NULL 0x0
#define CAPS_CONTACT 0x1
#define CAPS_TRANSPORT 0x2

#ifdef __cplusplus
extern "C" {
#endif
	void* caps_init(int version);
	void caps_deinit(void*c);
	void caps_add_identity(void*caps, const char *category, const char *type, const char *lang,const char *name);
	void caps_add_feature(void*caps, const char *feature);
	void caps_fill_info(void *caps, xmlnode x);
	void caps_fill_string(void *caps, char *dest, int len, const char *version);
	const char * caps_get_node(void *caps);
#ifdef __cplusplus
};
#endif

#ifdef __cplusplus

#include <string>
#include <vector>

class Identity
{
public:
	Identity() : lang(""), category(""), name(""), type("") {};
	std::string lang, category, name, type;

	friend bool operator< (const Identity &t, const Identity &id) {
		int r = t.category.compare(id.category);
		if(r<0)
			return true;
		if(r>0)
			return false;
		
		r = t.type.compare(id.type);
		if(r<0)
			return true;
		if(r>0)
			return false;
		
		r = t.lang.compare(id.lang);
		if(r<0)
			return true;
		if(r>0)
			return false;
		
		r = t.name.compare(id.name);
		if(r<0)
			return true;
		if(r>0)
			return false;
		return false;
	};
};

class Caps
{
public:
	Caps();
	virtual ~Caps();
	void init();

	std::string to_string(const char *version = NULL);
	void fill_info(xmlnode &query);
	
	std::string b64(std::string);
	void addIdentity(Identity i);
	void clearIdentities();
	
	void addFeature(std::string f);
	void clearFeatures();
	
	void setNode(std::string newnode);
	std::string &getNode();
private:
	std::vector<std::string> _features;
	std::vector<Identity> _identities;
	std::string _node, _cnode;
};
#endif

#endif /*CAPS_H_*/
