/*
 * SNAC - ICON
 *
 * Copyright (C) 2007 Grigory Holomiev <omever@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#ifndef SNAC_ICON_H
#define SNAC_ICON_H

#include <string>

#include <libicq2000/SNAC-base.h>

namespace ICQ2000 {

  // Buddy ICONS (Family 0x0010)
  const unsigned short SNAC_ICON_RequestError = 0x0001;
  const unsigned short SNAC_ICON_Request = 0x0006;
  const unsigned short SNAC_ICON_Response = 0x0007;

  // --------------------- Buddy icon (Family 0x0010) SNACs ---------

  class ICONFamilySNAC : virtual public SNAC {
   public:
    unsigned short Family() const { return SNAC_FAM_ICON; }
  };

  class ICONRequestSNAC : public ICONFamilySNAC, public OutSNAC {
   protected:
    unsigned int m_uin;
    unsigned short m_id;
    unsigned char m_hashlen, m_flags;
    unsigned char m_hash[16];
    
    
    void OutputBody(Buffer& b) const;

   public:
    ICONRequestSNAC(const unsigned int uin, unsigned short id, unsigned char flags, unsigned char hashlen, const unsigned char *md5);

    unsigned short Subtype() const { return SNAC_ICON_Request; }
   };
 
   class ICONRequestErrorSNAC : public ICONFamilySNAC, public InSNAC {
    protected:
     void ParseBody(Buffer& b);
     
    public:
     ICONRequestErrorSNAC();
     
     unsigned short Subtype() const { return SNAC_ICON_RequestError; }
  };

  class ICONResponseSNAC : public ICONFamilySNAC, public InSNAC {
   protected:
    unsigned int m_uin;
    unsigned short m_length;
    unsigned char * buf;
    
    void ParseBody(Buffer& b);
    
   public:
    ICONResponseSNAC();

    unsigned short Subtype() const { return SNAC_ICON_Response; }
  };

}

#endif
