#include "AvatarSpool.h"
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

using namespace std;

AvatarSpool::AvatarSpool(std::string rootdir)
{
    _rootdir = rootdir;
}

void AvatarSpool::storeBindata(std::vector<std::string> path, std::vector<unsigned char> bin)
{
    // Check and build path we going to use
    
    std::string filename (_rootdir);
    while(path.size()>1) {
	filename += "/";
	filename += path.at(0);
	path.erase(path.begin());
	
	if(access(filename.c_str(), F_OK) != 0) {
	    if(mkdir(filename.c_str(), 0777) ) {
		cout << "Cannot make dir " << filename << endl;
	     // Unable to make suitable path :(
		return;
	    }
	}
    }
    
    filename += "/" + path.at(0);
    ofstream fout(filename.c_str(), ios::binary);
    
    if(!fout) {
	cout << "Cannot open file " << filename << endl;
	return;
    }
    
    std::vector<unsigned char>::iterator i = bin.begin();
    while(i!=bin.end()) {
	fout.write((const char*)&(*i), 1);
	++i;
    }
    // All ok
}

std::vector<unsigned char> AvatarSpool::getBindata(std::vector<std::string> path)
{
    std::vector<unsigned char> ret;
    
    std::string filename (_rootdir);
    while(path.size()>0) {
	filename += "/";
	filename += path.at(0);
	path.erase(path.begin());
	
	if(access(filename.c_str(), F_OK) != 0) {
	    cout << "Cannot find path " << filename << endl;
	    return ret;
	}
    }
    
    ifstream fin(filename.c_str(), ios::in | ios::binary);

    char ch;
    while(fin) {
	fin >> ch;
	ret.push_back(ch);
    }
    
    return ret;
}

bool AvatarSpool::checkPath(std::vector<std::string> path)
{
    std::string filename (_rootdir);
    while(path.size()>0) {
	filename += "/";
	filename += path.at(0);
	path.erase(path.begin());
	
	if(access(filename.c_str(), F_OK) != 0) {
	    cout << "Cannot find path " << filename << endl;
		return false;
	}
    }
    return true;
}
