/*
 * Extended Presence Block
 *
 * Copyright (C) 2007 Grigory Holomiev <omever@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#include <libicq2000/PExtDataBlock.h>

#include <libicq2000/Contact.h>
#include <libicq2000/TLV.h>

using std::string;

namespace ICQ2000 {

  PExtDataBlock::PExtDataBlock()
  { 
      m_avatar_id = 0;
      m_avatar_hashlen = 0;
      m_avatar_flags = 0;
      memset(m_avatar_hash, 0, 16);
  }

  unsigned short PExtDataBlock::getAvatarId() const { return m_avatar_id; }
  unsigned char PExtDataBlock::getAvatarFlags() const { return m_avatar_flags; }
  unsigned char PExtDataBlock::getAvatarHashlen() const { return m_avatar_hashlen; }
  std::string PExtDataBlock::getMood() const { return m_mood; }
  std::string PExtDataBlock::getMoodIcon() const { return m_moodicon; }
  std::string PExtDataBlock::getTune() const { return m_tune; }
  const unsigned char * PExtDataBlock::getAvatarHash() const { return m_avatar_hash; }

  void PExtDataBlock::Parse(Buffer& b) {

    std::string m_screenname;

    b.UnpackByteString(m_screenname);
    unsigned short m_warninglevel;
    
printf("ScreenName: %s\n", m_screenname.c_str());
    
    b >> m_warninglevel;
    unsigned short no_tlvs;
    b >> no_tlvs;
    
    TLVList tlvlist;
    tlvlist.Parse(b, TLV_ParseMode_InAdvPresenceData, no_tlvs);

    if (tlvlist.exists(TLV_Extdata)) {
      printf("he is got an extinfo!\n");
      ExtDataTLV *t = (ExtDataTLV*)tlvlist[TLV_Extdata];
      m_avatar_id = t->getId();
      m_avatar_hashlen = t->getHashlen();
      m_avatar_flags = t->getFlags();
      memcpy(m_avatar_hash, t->getHash(), 16);
      m_mood = t->getMood();
 printf("MOODD: %s\n", m_mood.c_str());
      m_moodicon = t->getMoodIcon();
      m_tune = t->getTune();
    }

  }

}
