/*
 * SNAC - Buddy ICON services
 *
 * Copyright (C) 2007 Grigory Holomiev <omever@gmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
 *
 */

#include <libicq2000/SNAC-ICON.h>

#include <libicq2000/buffer.h>
#include <libicq2000/Contact.h>
#include <string.h>

using std::string;

namespace ICQ2000 {

  // -------------- Buddy ICON (0x0010) Family -----------------------

  ICONRequestSNAC::ICONRequestSNAC(const unsigned int uin, unsigned short id, unsigned char flags, unsigned char hashlen, const unsigned char *md5)
  {
    m_uin = uin;
    m_id = id;
    m_flags = flags;
    m_hashlen = hashlen;
    memcpy(m_hash, md5, 16);
  }

  void ICONRequestSNAC::OutputBody(Buffer& b) const{
    std::string uin = Contact::UINtoString(m_uin);
    std::cout << "ICONRequestSNAC" << std::endl;
    b<<(unsigned char)uin.length();
    b.setLittleEndian();
    b.PackUint16StringNull(uin);
    b.setBigEndian();

    b<<(unsigned char)0x01; // unknown (command?)
    b<<(unsigned short)0x0001; // icon id
    b<<(unsigned char)0x01;  // icon flags
    b<<(unsigned char)0x10;  // md5 hash len
    b.Pack(m_hash, 16);

    b.dump(std::cout);
  }

  ICONResponseSNAC::ICONResponseSNAC() { }

  void ICONResponseSNAC::ParseBody(Buffer& b){
    std::cout << "ICONResponseSNAC" << std::endl;
    b.dump(std::cout);
    b.advance(b.size());
  }

  ICONRequestErrorSNAC::ICONRequestErrorSNAC() { }

  void ICONRequestErrorSNAC::ParseBody(Buffer& b){
    std::cout << "ICONRequestErrorSNAC" << std::endl;
    b.dump(std::cout);
//    b.advance(32);
  }

}
