/*
 * SNAC - Server-based lists
 * Mitz Pettel, 2001
 *
 * based on: SNAC - Buddy List
 *
 * Copyright (C) 2001 Barnaby Gray <barnaby@beedesign.co.uk>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include <libicq2000/SNAC-SBL.h>

#include <libicq2000/TLV.h>

using std::string;

namespace ICQ2000 {

  // --------------- Server-based Lists (Family 0x0013) SNACs --------------

  RequestSBLSNAC::RequestSBLSNAC() { }

  void RequestSBLSNAC::OutputBody(Buffer& b) const {
     printf("Request SBLSNAC\n");
//      b << (unsigned short)(0x13);
//      b << (unsigned short)(0x4);

//    b << (unsigned short)0x0000;

    // a request id
//    b << (unsigned short)m_requestID; /* low word of the request ID */

      b.dump(std::cout);
    }

  ActivateSBLSNAC::ActivateSBLSNAC() { }

  void ActivateSBLSNAC::OutputBody(Buffer& b) const {
     printf("Activate SBLSNAC\n");
//      b << (unsigned short)(0x13);
//      b << (unsigned short)(0x7);

//    b << (unsigned short)0x0000;

    // a request id
//    b << (unsigned short)m_requestID; /* low word of the request ID */

    }

  SBLListSNAC::SBLListSNAC() { }

  void SBLListSNAC::ParseBody(Buffer& b) {
        std::cout << "Buffer with SBL: " << std::endl << b;
        b.advance(1);			// 00
        unsigned short entityCount;
        b >> entityCount;		// number of entities?
	unsigned short ic = 0;
        while (b.pos()<=b.size()-10)
        {
            unsigned short nameLength;
	    unsigned short m_groupid, m_itemid, m_itemType;

            b >> nameLength;
            string name;
            b.Unpack(name, nameLength);


	    b >> m_groupid;
	    b >> m_itemid;

	    b >> m_itemType;
            unsigned short dataLength;
            b >> dataLength;

	    switch (m_itemType)
	    {
	    default:
		b.advance(dataLength);
		break;
            case 0x0: // Buddy record
	    {
		ContactRef c(new Contact(Contact::StringtoUIN(name)));
		c->setGroupID(m_groupid);
		c->setItemID(m_itemid);
        	m_contacts.add(c);

        	while (dataLength>=4)
        	{
		    unsigned short infoType;
		    unsigned short infoLength;
		    b >> infoType;
		    dataLength -= 2;
		    b >> infoLength;
			dataLength -= 2;
			ic++;
			string tmp;
			tmp = "";
			printf("I've got: %X:%X\n", infoType, infoLength);
			switch(infoType)
			{
			    case 0x0131: // UIN
				if(infoLength) {
				    b.Unpack(tmp, infoLength);
				}
				c->setAlias(tmp);
				break;
			    case 0x0066: // Not authorized yet
			    c->setAuthwaiting(true);
				break;
			    case 0x0137: // Assigned mail address
				if(infoLength) {
				    b.Unpack(tmp, infoLength);
				}
				c->setEmail(tmp);
				break;
			    case 0x013a: // Assigned sms number
				if(infoLength) {
				    b.Unpack(tmp, infoLength);
				}
				c->setMobileNo(tmp);
				break;
			    case 0x013c: // Assigned comment for buddy
				if(infoLength) {
				    b.Unpack(tmp, infoLength);
				}
				c->setAssignedComment(tmp);
				break;
			    default:	// other stuff we don't understand
				b.advance(infoLength);
				break;
			}
			dataLength -= infoLength;
        	}
        	b.advance(dataLength);
		break;
	    }
	    case 0x1: // Group record
		m_groups[name] = m_groupid;
		b.advance(dataLength);
		break;
    	    }
	}
//	b.dump(std::cout);
	unsigned int ts;
	b >> ts;

	printf("Current ts: %d, ec: %d\n", ts, ic);

	m_contacts.setEC(ic);

	if(ts)
	    m_contacts.setLastChanged(ts);
//        b.advance(4);
    }

  SBLParamsListSNAC::SBLParamsListSNAC() { }

  void SBLParamsListSNAC::ParseBody(Buffer& b) {

        b.advance(b.size());
    }

  SBLParamsSNAC::SBLParamsSNAC() {};
  void SBLParamsSNAC::OutputBody(Buffer& b) const {
  }

  SBCLSyncDoneSNAC::SBCLSyncDoneSNAC(unsigned int ts, unsigned short count) :m_ts(ts), m_count(count) {};

  void SBCLSyncDoneSNAC::OutputBody(Buffer& b) const {
    // this is a TLV header
//    b << (unsigned short)0x0013
//      << (unsigned short)0x0005;

//    b << (unsigned short)0x0000;

    // a request id
//    b << (unsigned int)m_requestID; /* low word of the request ID */

    b << (unsigned int) m_ts; /* ts of the contact-list */
    b << (unsigned short) m_count; /* number of users in CL */

    printf("i'm sending TS=%d, CNT=%d\n", m_ts, m_count);
  }

  void SBL_SSI_Add_Item::OutputBody(Buffer& b) const {

    b << m_uin;
    b << m_groupid;
    b << m_itemid;

    // Type of record
    b << (unsigned short) 0x0000; // Buddy

    // Additional data length
    if(m_req) {
    	b << (unsigned short) 0x0004;
    	b << (unsigned short) 0x0066;
    	b << (unsigned short) 0x0000;
    } else {
    	b << (unsigned short) 0x0000;
    }
  }

  void SBL_SSI_Add_Group::OutputBody(Buffer& b) const {

    b << m_group;
    b << m_groupid;
    b << (unsigned short)0;

    // Type of record
    b << (unsigned short) 0x0001; // Group

    // Additional data length
    b << (unsigned short) 0x0000;
  }

  void SBL_SSI_Remove_Item::OutputBody(Buffer& b) const {
    b << m_uin;
    b << m_groupid;
    b << m_itemid;

    // Type of record
    b << (unsigned short) 0x0000; // Buddy

    // Additional data length
    b << (unsigned short) 0x0000;
  }

  void SBL_SSI_Edit_Ack::ParseBody(Buffer &b) {
    if(m_codes.size())
	m_codes.clear();
    while (b.pos()<b.size())
    {
	unsigned short tmp;
	b >> tmp;
	m_codes.push_back(tmp);
    }
  }

  void SBL_Auth_Request_Send::OutputBody(Buffer& b) const {
    char tmp[128];
    snprintf(tmp, 127, "%d", m_uin);

    b.PackByteString(tmp);
    b << m_reason;
    b << (unsigned short) 0x0;
  }

  void SBL_Auth_Reply_Ack::ParseBody(Buffer &b) {
    b >> m_uin;
    unsigned char tmp;
    b >> tmp;
    m_accepted = (tmp)?true:false;
    b >> m_reason;
  }

  void SBL_FutureAuth_Grant::OutputBody(Buffer& b) const {
    char tmp[128];
    snprintf(tmp, 127, "%d", m_uin);

    b.PackByteString(std::string(tmp));
    b << m_reason;
    b << (unsigned short) 0x0;
  }

  void SBL_RemoveMe_SNAC::OutputBody(Buffer& b) const {
    b.PackByteString(m_uin);
  }
}
