#ifndef BINVECTOR_H
#define BINVECTOR_H

/* Simple binary data vector implementation
   Author: Grigory Holomiev (omever@gmail.com) 2007
 
   Covered by GNU General Public License
*/

#include <vector>
#include <cstdio>

class binvector : public std::vector<unsigned char>
{
public:
	binvector() {
		_tmpdata = NULL;
	}
	
	binvector(unsigned char * in, unsigned int len) {
		_tmpdata = NULL;
		
		for(unsigned int i=0; i<len; ++i)
			this->push_back(in[i]);
	}

	~binvector()
	{
		if(_tmpdata != NULL) {
			delete [] _tmpdata;
		}
	}

	const unsigned char* cdata()
	{
		if(_tmpdata != NULL) {
			delete [] _tmpdata;
		}

		unsigned int l = size();
		_tmpdata = new unsigned char[l];
		
		for(unsigned int i=0; i<l; ++i)
			_tmpdata[i] = at(i);

		return _tmpdata;
	}

	const unsigned char* c_str()
	{
		if(_tmpdata != NULL) {
			delete [] _tmpdata;
		}

		unsigned int l = 3*(size()+1); // 3 ascii bytes per binary byte (2hex and 1space) and a trailing 0 byte plus leading '0x'
		
		_tmpdata = new unsigned char[l];
		
		sprintf((char*)_tmpdata, "0x");
		for(unsigned int i=0; i<size(); ++i) {
			sprintf((char*)(_tmpdata+2+i*3), "%2x ", at(i) );
		}
		_tmpdata[l-1] = 0;
		
		return _tmpdata;
	}


private:
	unsigned char *_tmpdata;
};

    
#endif

