#ifndef PQUEUE_H
#define PQUEUE_H

#include <queue>
#include <pthread.h>

/* PThreaded queue by Grigory Holomiev (omever@ya.ru) released under terms of GNU GPL v2 */
template <class T> class TQueue
{
public:
    TQueue() 
    {
	pthread_mutex_init(&__mutex, NULL);
    }

    ~TQueue() 
    {
	pthread_mutex_destroy(&__mutex);
    }
    
    unsigned int size()
    {
        unsigned int tmp;
	_lock();
        tmp = __queue.size();
	_unlock();
	return tmp;
    }
    
    T front()
    {
        T tmp;
        _lock();
        tmp = __queue.front();
        _unlock();
        return tmp;
    }
    
    void pop()
    {
        _lock();
        __queue.pop();
        _unlock();
    }
    
    void push(T in)
    {
	_lock();
        __queue.push(in);
        _unlock();
    }
    
protected:
    void _lock() 
    {
        pthread_mutex_lock(&__mutex);
    }
    
    void _unlock()
    {
        pthread_mutex_unlock(&__mutex);
    }
    
private:
    pthread_mutex_t __mutex;
    std::queue<T> __queue;
};

#endif
